#!/usr/bin/env python3
# Copyright 2004-present Facebook. All Rights Reserved.

from typing import Optional


class AioMsaError(Exception):
    """Base class for all *aiomsa* errors."""

    pass


class DuplicateRouteError(AioMsaError):
    """Raised if more than one route has the same method and path."""

    pass


class ClientError(AioMsaError):
    """Base class for all client errors."""

    pass


class ClientStoppedError(ClientError):  # pragma: no cover
    """Raised if a client is used before it's started.

    Args:
        msg: The exception message.
    """

    def __init__(self, msg: Optional[str] = None) -> None:
        if msg is None:
            msg = "The client cannot be used before it's started"
        super().__init__(msg)


class ClientRuntimeError(ClientError):  # pragma: no cover
    """Raised if a client operation fails.

    Args:
        msg: The exception message.
    """

    def __init__(self, msg: Optional[str] = None) -> None:
        if msg is None:
            msg = "An issue with the client occurred at runtime"
        super().__init__(msg)



#!/usr/bin/env python3
# Copyright 2004-present Facebook. All Rights Reserved.

import asyncio
import base64
import itertools
import json
from typing import Dict, List, Optional, Tuple

from ... import abc


class Message:
    """Takes base64 encoded header and message and stores them as bytes."""

    def __init__(self, header: str, message: str, delay_ms: int = 0) -> None:
        self.header = base64.b64decode(header)
        self.message = base64.b64decode(message)
        self.delay_ms = delay_ms

    def dump(self) -> Tuple[bytes, bytes]:
        return self.header, self.message


class Stream:
    """A set of messages emitted by this module repeatedly."""

    def __init__(
        self,
        messages: List[Dict],
        repeat: int = 1,
    ) -> None:
        self.repeat = repeat
        self.messages = [Message(**message) for message in messages]


class StaticNode(abc.E2Node):
    """A mock E2Node for E2 message playback."""

    def __init__(self, e2_node_id: str, streams: List[Stream]) -> None:
        super().__init__(e2_node_id, [])
        self.streams = streams


class Subscription(abc.Subscription):
    def __init__(self, node: StaticNode) -> None:
        self.node = node
        self.message_iter = itertools.chain.from_iterable(
            [
                itertools.chain.from_iterable(
                    itertools.repeat(stream.messages, stream.repeat)
                )
                for stream in node.streams
            ]
        )

    @property
    def id(self) -> str:
        return self.node.id

    def __aiter__(self) -> "Subscription":
        return self

    async def __anext__(self) -> Tuple[bytes, bytes]:
        while True:
            try:
                message = next(self.message_iter)
                await asyncio.sleep(message.delay_ms / 1000)
                return message.dump()
            except StopIteration:
                raise StopAsyncIteration


class E2Client(abc.E2Client):
    def __init__(self, staticfile: str) -> None:
        self.e2nodes = {}

        with open(staticfile) as f:
            static_data = json.loads(f.read())

        for k, v in static_data["nodes"].items():
            streams = [Stream(**s) for s in v["streams"]]
            self.e2nodes[k] = StaticNode(k, streams)

    async def list_nodes(self, oid: Optional[str] = None) -> List[abc.E2Node]:
        return list(self.e2nodes.values())

    async def control(
        self,
        e2_node_id: str,
        service_model_name: str,
        service_model_version: str,
        header: bytes,
        message: bytes,
        control_ack_request: abc.RICControlAckRequest,
    ) -> Optional[bytes]:
        pass

    async def subscribe(
        self,
        e2_node_id: str,
        service_model_name: str,
        service_model_version: str,
        subscription_id: str,
        trigger: bytes,
        actions: List[abc.RICAction],
    ) -> Subscription:
        return Subscription(self.e2nodes[e2_node_id])

    async def unsubscribe(
        self,
        e2_node_id: str,
        service_model_name: str,
        service_model_version: str,
        subscription_id: str,
    ) -> None:
        pass

    async def __aenter__(self) -> "E2Client":
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        pass



#!/usr/bin/env python3
# Copyright 2004-present Facebook. All Rights Reserved.

import base64
import json
import pathlib
from collections import deque
from typing import Any, Dict, List, Tuple

from ..abc import Subscription


class RecordSubscription:
    """A wrapper class for recording E2 messages from aiomsa.abc.Subscription
    types. When passed to the RecordSubscription constructor, instances of
    aiomsa.abc.Subscription are wrapped and the returned RecordSubscription
    instance may be used identically to the original Subscription instance. The
    RecordSubscription instance has additional functionality: the ability to
    dump the entire history of E2 messages received to a dict or JSON string.

    Messages are stored in a FIFO queue, and old messages are dequeued when the
    queue reaches maxlen.

    Args:
        sub: An instance of a aiomsa.abc.Subscription derived class.
        maxlen: The maximum number of E2 messages stored.
    """

    ATTRS = [
        "_messages",
        "_sub",
    ]

    def __init__(self, sub: Subscription, maxlen: int = 200) -> None:
        self._sub = sub
        self._messages = deque([], maxlen)

    async def __anext__(self) -> Tuple[bytes, bytes]:
        ret = await self._sub.__anext__()
        self._record(*ret)
        return ret

    def __aiter__(self) -> "RecordSubscription":
        return self

    def _record(self, header: bytes, message: bytes) -> None:
        """Save an E2 message.

        Args:
            header: E2 message header
            message: E2 message body
        """
        self._messages.append(
            [
                {
                    "header": base64.b64encode(header).decode("ascii"),
                    "message": base64.b64encode(message).decode("ascii"),
                }
            ]
        )

    def to_dict(self) -> Dict[str, Any]:
        """Dump recorded messages to a dict comprehensible to aiomsa.mock.

        Returns:
            A dict containing a "node" object comprehensible to aiomsa.mock.e2.
        """
        return {
            "e2_node_id": self.id,
            "streams": [{"messages": list(self._messages)}],
        }

    def dump(self) -> str:
        """Dump recorded messages to a JSON comprehensible to aiomsa.mock.

        Returns:
            A JSON string of a "node" object comprehensible to aiomsa.mock.e2.
        """
        json_str = json.dumps(self.to_dict())
        self._messages.clear()
        return json_str

    def __getattr__(self, attr: str) -> Any:
        if attr not in self.ATTRS:
            return getattr(self._sub, attr)
        else:
            return self.__dict__[attr]

    def __setattr__(self, attr: str, val: Any) -> None:
        if attr not in self.ATTRS:
            return setattr(self._sub, attr, val)
        else:
            self.__dict__[attr] = val


def generate_staticfile(filenames: List[str], output: str) -> None:
    """Join multiple JSONs dumped from RecordSubscription instances to a
    staticfile for aiomsa.mock.e2.static.

    Args:
        filenames: JSONs generated by RecordSubscription.dump
        output: Output file name for JSON created by the joined files.
    """
    staticdict = {"nodes": []}
    for f in filenames:
        node = json.loads(pathlib.Path(f).read_text())
        staticdict["nodes"] += [node]
    with open(output, "w") as f:
        f.write(json.dumps(staticdict))



#!/usr/bin/env python3
# Copyright 2004-present Facebook. All Rights Reserved.

from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import IntEnum
from types import TracebackType
from typing import AsyncIterator, List, Optional, Tuple, Type


class RICActionType(IntEnum):
    """An :class:`~.enum.IntEnum` defining possible RIC actions.

    Attributes:
        REPORT
        INSERT
        POLICY
    """

    REPORT = 0
    INSERT = 1
    POLICY = 2


class RICControlAckRequest(IntEnum):
    """An :class:`~.enum.IntEnum` defining circumstances in which the E2 node
    can/should reply to a RIC control acknowledge message.

    Attributes:
        ACK
        NO_ACK
        NACK
    """

    ACK = 0
    NO_ACK = 1
    NACK = 2


class RICSubsequentActionType(IntEnum):
    """An :class:`~.enum.IntEnum` defining the valid actions that can be taken after
    completing a particular :class:`~.RICAction`.

    Attributes:
        CONTINUE
        WAIT
    """

    CONTINUE = 0
    WAIT = 1


class RICTimeToWait(IntEnum):
    """An :class:`~.enum.IntEnum` defining the time to wait after completing a
    particular :class:`~.RICAction`.

    Attributes:
        ZERO
        W1MS
        W2MS
        W5MS
        W10MS
        W20MS
        W30MS
        W40MS
        W50MS
        W100MS
        W200MS
        W500MS
        W1S
        W2S
        W5S
        W10S
        W20S
        W60S
    """

    ZERO = 0
    W1MS = 1
    W2MS = 2
    W5MS = 3
    W10MS = 4
    W20MS = 5
    W30MS = 6
    W40MS = 7
    W50MS = 8
    W100MS = 9
    W200MS = 10
    W500MS = 11
    W1S = 12
    W2S = 13
    W5S = 14
    W10S = 15
    W20S = 16
    W60S = 17


@dataclass
class RICSubsequentAction:
    """The subsequent action to take once the action is complete.

    Args:
        type: The subsequent action type.
        time_to_wait: Time to wait before performing the subsequent action.
    """

    type: RICSubsequentActionType
    time_to_wait: RICTimeToWait


@dataclass
class RICAction:
    """An action to be taken in a :meth:`~.E2Client.subscribe` request.

    Args:
        id: The RIC action ID.
        type: The action type to be executed.
        subsequent_action: The subsequent action to take once the action is complete.
        definition: Parameters used when executing a report, insert, or policy service.
    """

    id: int
    type: RICActionType
    subsequent_action: Optional[RICSubsequentAction] = None
    definition: Optional[bytes] = None

    def __post_init__(self) -> None:
        if self.type == RICActionType.INSERT and self.subsequent_action is None:
            raise ValueError(
                "subsequent_action must be present when RICActionType is set to 'INSERT'"
            )


@dataclass
class ReportStyle:
    """A RIC report style.

    Args:
        type: The report style type.
        name: The report style name.
    """

    type: int
    name: str


@dataclass
class KpmReportStyle(ReportStyle):
    """A RIC KPM report style.

    Args:
        measurements: The list of KPM measurements in tuple form: (ID, name).
    """

    measurements: List[Tuple[int, str]]


@dataclass
class RanFunction:
    """A RAN function belonging to a specific service model.

    Args:
        id: The RAN function ID.
        report_styles: The list of supported report styles.
    """

    id: int
    report_styles: List[ReportStyle]


@dataclass
class ServiceModel:
    """A service model belonging to an E2 node.

    Args:
        name: The service model name.
        oid: The service model object ID.
        ran_functions: The list of supported RAN functions.
    """

    name: str
    oid: str
    ran_functions: List[RanFunction]


@dataclass
class E2Cell:
    """A cell belonging to an E2 node.

    Args:
        id: The cell global ID.
        oid: The cell object ID.
        pci: The cell physical ID.
    """

    id: str
    oid: str
    pci: Optional[int] = None


@dataclass
class E2Node:
    """A logical node terminating E2 interface.

    Args:
        id: The E2 node ID.
        service_models: The list of supported service models.
    """

    id: str
    service_models: List[ServiceModel]


class Subscription(ABC):
    @property
    @abstractmethod
    def id(self) -> str:
        """Return an identifier for the subscription."""
        pass

    @abstractmethod
    def __aiter__(self) -> "Subscription":
        """Initialize any resources, if needed."""
        pass

    @abstractmethod
    async def __anext__(self) -> Tuple[bytes, bytes]:
        """Return the next indication in the subscription.

        Returns:
            The next indication header and message, if available.

        Raises:
            StopAsyncIteration: The subscription has been exhausted.
        """
        pass


class E2Client(ABC):
    @abstractmethod
    async def control(
        self,
        e2_node_id: str,
        service_model_name: str,
        service_model_version: str,
        header: bytes,
        message: bytes,
        control_ack_request: RICControlAckRequest,
    ) -> Optional[bytes]:
        """Send a control message to the RIC to initiate or resume some functionality.

        Args:
            e2_node_id: The target E2 node ID.
            service_model_name: The service model name.
            service_model_version: The service model version.
            header: The RIC control header.
            message: The RIC control message.
            control_ack_request: Instruct whether/how the E2 node should reply.

        Returns:
            The control outcome, if specifically requested via ``control_ack_request``,
            else ``None``.

        Raises:
            ClientStoppedError: The underlying client resources have not been started.
            ClientRuntimeError: There was an error performing the request.
        """
        pass

    @abstractmethod
    async def subscribe(
        self,
        e2_node_id: str,
        service_model_name: str,
        service_model_version: str,
        subscription_id: str,
        trigger: bytes,
        actions: List[RICAction],
    ) -> Subscription:
        """Establish an E2 subscription.

        Args:
            e2_node_id: The target E2 node ID.
            service_model_name: The service model name.
            service_model_version: The service model version.
            subscription_id: The ID to use for the subscription.
            trigger: The event trigger.
            actions: A sequence of RIC service actions.

        Returns:
            The created subscription.

        Raises:
            ClientStoppedError: The underlying client resources have not been started.
            ClientRuntimeError: There was an error performing the request.
        """
        pass

    @abstractmethod
    async def unsubscribe(
        self,
        e2_node_id: str,
        service_model_name: str,
        service_model_vesrion: str,
        subscription_id: str,
    ) -> None:
        """Delete an E2 subscription.

        Args:
            e2_node_id: The target E2 node ID.
            service_model_name: The service model name.
            service_model_version: The service model version.
            subscription_id: The ID of the subscription to delete.

        Raises:
            ClientStoppedError: The underlying client resources have not been started.
            ClientRuntimeError: There was an error performing the request.
        """
        pass

    @abstractmethod
    async def __aenter__(self) -> "E2Client":
        """Create any underlying resources required for the client to run."""
        pass

    @abstractmethod
    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        """Cleanly stop all underlying resources used by the client."""
        pass


class SDLClient(ABC):
    @abstractmethod
    async def get_cells(self, e2_node_id: str) -> List[E2Cell]:
        """Get the cells corresponding to the given E2 node ID.

        Args:
            e2_node_id: The target E2 node ID.

        Returns:
            A list of cells that belong to ``e2_node_id``.

        Raises:
            ClientStoppedError: The underlying client resources have not been started.
            ClientRuntimeError: There was an error performing the request.
        """
        pass

    @abstractmethod
    def watch_e2_connections(self) -> AsyncIterator[E2Node]:
        """Stream for newly available E2 node connections.

        Yields:
            An available :class:`E2Node` object.

        Raises:
            ClientStoppedError: The underlying client resources have not been started.
            ClientRuntimeError: There was an error performing the request.
        """
        pass

    @abstractmethod
    async def __aenter__(self) -> "SDLClient":
        """Create any underlying resources required for the client to run."""
        pass

    @abstractmethod
    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        """Cleanly stop all underlying resources used by the client."""
        pass



#!/usr/bin/env python3
# Copyright 2004-present Facebook. All Rights Reserved.

__all__ = ["run"]

import asyncio
import logging
import os
import signal
import traceback
from typing import Coroutine, Optional

from aiohttp import web
from aiohttp_swagger import setup_swagger  # type: ignore

from .exceptions import DuplicateRouteError
from .server import error_middleware, routes


def run(
    main: Coroutine,
    path: str,
    app: Optional[web.Application] = None,
    **server_kwargs,
) -> None:
    """Start the webserver and the entrypoint logic passed in as ``main``.

    Args:
        main: The entrypoint for the service's logic, in the form of a coroutine.
        path: The path to the service's configuration file on disk.
        app: An existing web application object, if available.
        server_kwargs: Variable number of ``kwargs`` to pass to
                       :func:`aiohttp.web.run_app`.

    Raises:
        DuplicateRouteError: A user-supplied route conflicts with one of the default
                             :doc:`routes<./routes>`.
        ValueError: ``main`` is not a proper coroutine.
    """
    logging.basicConfig(
        format="%(levelname)s %(asctime)s %(filename)s:%(lineno)d] %(message)s",
        level=logging.INFO,
        datefmt="%Y-%m-%d %H:%M:%S",
    )

    if not asyncio.coroutines.iscoroutine(main):
        raise ValueError(f"A coroutine was expected, got {main}")

    # Create web application object and shutdown event
    if app is None:
        app = web.Application(middlewares=[error_middleware])
    app["main"] = main
    app["path"] = path
    app["shutdown_event"] = asyncio.Event()

    # Initialize the endpoints for the HTTP server
    try:
        app.add_routes(routes)
    except RuntimeError:
        app["main"].close()
        resources = [(r.method, r.path) for r in routes if isinstance(r, web.RouteDef)]
        raise DuplicateRouteError(
            f"A user-supplied route conflicts with a pre-registered route: {resources}"
        )

    # Document the endpoints with OpenAPI
    setup_swagger(app, ui_version=3)

    # Add background tasks
    app.on_startup.append(on_startup)
    app.on_cleanup.append(on_cleanup)
    web.run_app(app, **server_kwargs)


async def on_startup(app: web.Application) -> None:
    """Create the 'main_wrapper_task' and 'shutdown_listener'."""
    app["shutdown_listener"] = asyncio.create_task(shutdown_listener(app))
    app["main_wrapper_task"] = asyncio.create_task(main_wrapper(app))


async def on_cleanup(app: web.Application) -> None:
    """Cancel the 'shutdown_listener' and 'main_wrapper_task'."""
    app["shutdown_listener"].cancel()
    app["main_wrapper_task"].cancel()
    await app["shutdown_listener"]

    try:
        await app["main_wrapper_task"]
    except:  # noqa: E722
        traceback.print_exc(chain=False)


async def main_wrapper(app: web.Application) -> None:
    """Run the 'main' coroutine and set the 'shutdown_event' if it fails."""
    try:
        await app["main"]
    except asyncio.CancelledError:
        pass
    except:  # noqa: E722
        app["shutdown_event"].set()
        raise


async def shutdown_listener(app: web.Application) -> None:
    """Wait for the 'shutdown_event' notification to kill the process."""
    try:
        await app["shutdown_event"].wait()
        logging.warning("Shutting down!")

        # Wait before shutting down
        await asyncio.sleep(1)
    except asyncio.CancelledError:
        pass
    finally:
        os.kill(os.getpid(), signal.SIGTERM)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['aiomsa', 'aiomsa.mock', 'aiomsa.mock.e2', 'aiomsa.utils']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp-swagger>=1.0.9,<2.0.0',
 'aiohttp>=3.5.4,<4.0.0',
 'prometheus-async>=19.2.0,<20.0.0']

extras_require = \
{'docs': ['furo>=2021.2.21b25,<2022.0.0',
          'sphinx>=3.4.3,<4.0.0',
          'sphinx-autodoc-typehints>=1.11.1,<2.0.0',
          'sphinxcontrib-openapi>=0.7.0,<0.8.0']}

setup_kwargs = {
    'name': 'aiomsa',
    'version': '0.1.0a8',
    'description': 'Asynchronous xApp framework',
    'long_description': '# aiomsa\n[![build](https://github.com/facebookexternal/aiomsa/actions/workflows/build.yml/badge.svg)](https://github.com/facebookexternal/aiomsa/actions/workflows/build.yml)\n[![style](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n![PyPI - Downloads](https://img.shields.io/pypi/dw/aiomsa)\n\n*aiomsa* is a Python 3.7+ framework built using `asyncio`. At its core, *aiomsa*\nprovides a simple and standardized way to write xApps that can be deployed as\nmicroservices in Python.\n\n## Installation\n*aiomsa* can be installed from PyPI.\n```bash\npip install aiomsa\n```\n\nYou can also get the latest code from GitHub.\n```bash\npoetry add git+https://github.com/facebookexternal/aiomsa\n```\n\n## Getting Started\nThe follwing example shows how to use *aiomsa* to create a simple xApp for subscribing\nto the E2T service for a particular custom service model.\n\n```python\nimport asyncio\n\nimport aiomsa\nimport aiomsa.abc\nfrom onos_ric_sdk_py import E2Client, SDLClient\n\nfrom .models import MyModel\n\n\nasync def run(e2: aiomsa.abc.E2Client, e2_node_id: str) -> None:\n   subscription = await e2.subscribe(\n      e2_node_id,\n      service_model_name="my_model",\n      service_model_version="v1",\n      subscription_id="my_app-my_model-sub",\n      trigger=bytes(MyModel(param="foo")),\n      actions=[\n         aiomsa.abc.RICAction(\n            id=1,\n            type=aiomsa.abc.RICActionType.REPORT,\n            subsequent_action_type=aiomsa.abc.RICSubsequentActionType.CONTINUE,\n            time_to_wait=aiomsa.abc.RICTimeToWait.ZERO,\n         )\n      ],\n   )\n\n   async for (_header, message) in subscription:\n      print(message)\n\n\nasync def main() -> None:\n   async with E2Client(app_id="my_app", e2t_endpoint="e2t:5150") as e2, SDLClient(\n      topo_endpoint="topo:5150"\n   ) as sdl:\n      async for e2_node in sdl.watch_e2_connections():\n         asyncio.create_task(run(e2, e2_node.id))\n\n\nif __name__ == "__main__":\n   aiomsa.run(main())\n```\n',
    'author': 'Facebook Connectivity',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/facebookexternal/aiomsa',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
